// Copyright 1992-2007 by Jon Dart. All Rights Reserved.
//
#ifndef _MOVE_GENERATOR_H
#define _MOVE_GENERATOR_H

#include "board.h"
#include "constant.h"
#include "searchc.h"

class Move_Generator
{
    public:
		
    enum Phase {HASH_MOVE_PHASE, WINNING_CAPTURE_PHASE,
      KILLER1_PHASE, KILLER2_PHASE, HISTORY_PHASE,
      LOSERS_PHASE, LAST_PHASE };

    static void init();
    static void cleanup();

    Move_Generator( const Board &ABoard, 
                    SearchContext *,
                    unsigned ply, 
                    Move pvMove, 
                    int trace);
	
    // Generate the next move, in sorted order, NullMove if none left
    Move NextMove() {
       if (index >= batch_count) {
          if ((batch_count = getBatch(batch,index)) == 0) 
             return NullMove;
       }
       return batch[index++];
    }

    // Generate the next check evasion, NullMove if none left
    Move NextEvasion();

    // Generate only non-capturing moves.
    int GenerateNonCaptures(Move *moves);

    // fill array "moves" with moves, starting at "index", returns
    // # of moves (0 if no more). This tries to generate moves in stages,
    // so it should be called repeatedly until the return value is zero.
    // 		
    // The moves returned are generally "pseudo-legal", i.e. they may
    // involve moves into check.  However, if the side to move is in
    // check, then all moves returned are strictly legal.		
    //
    int GenerateAllMoves(Move *moves,int repeatable);

    // Like Generate_Moves, but only generates captures, promotions,
    // and check evasions.
    //
    int GenerateCaptures( Move *moves );
		
    // returns a list of the moves that may be used to escape
    // check.  Unlike the regular move generation routine, all
    // moves are checked for legality.
    //
    unsigned GenerateEvasions(Move * moves, int captures_only);

    // Generate non-capturing checking moves
    unsigned GenerateChecks(Move * moves);

    unsigned Moves_Generated()
    {
       return moves_generated;
    }

    Phase getPhase() const {
      return (Phase)phase;
    }
		
private:
		
    int getBatch(Move *&batch,int &index);
    int OrderWinningCaptures (int captures, Move hashMove);

    unsigned GenerateEvasionsCaptures(Move * moves);
    unsigned GenerateEvasionsNonCaptures(Move * moves);

    const Board &board;
    SearchContext *context;
    int ply;
    int moves_generated;
    int phase, winners, losers_count,index,batch_count,forced;
    Move hashMove;
    int non_captures_count,history_count;
    Bitmap king_attacks; // for evasions
    int num_attacks; // for evasions
    Square source; // for evasions;
    Move *batch;
    Move losers[100];
    Move moves[Constants::MaxMoves];
    Move killer1,killer2,killer3,killer4;
    int master;
        
    inline void SetMove( const Square source, const Square dest,
       const PieceType promotion,
       Move *moves, unsigned &NumMoves);

};

#endif
