// Copyright 2005-2007 by Jon Dart. All Rights Reserved.

#include "threadp.h"
#include "search.h"
#include <iostream>
#ifndef _WIN32
#include <errno.h>
#include <fcntl.h>
#endif
using namespace std;

#ifdef _WIN32
static DWORD WINAPI parkingLot(void *x)
#else
static void * CDECL parkingLot(void *x)
#endif
{
   ThreadInfo *ti = (ThreadInfo*)x;
   while (ti->state != ThreadInfo::Terminating) {
#ifdef _THREAD_TRACE
      cout << "thread in parkingLot" << endl;
#endif
      ti->state = ThreadInfo::Idle; // mark thread available again
      if (ti->wait()) break;
#ifdef _THREAD_TRACE
      cout <<"thread " << ti->index << " unblocked" << endl;
#endif
      if (ti->state == ThreadInfo::Terminating) break;
      //#ifdef _THREAD_TRACE
      //      cout << "calling searchSMP" << endl;
      //#endif
      ASSERT(ti->work);
      ti->work->searchSMP(ti);  
      ti->pool->checkIn(ti);
   }
   return 0;
}

ThreadInfo::~ThreadInfo() {
#ifdef _WIN32
  CloseHandle(hEvent1);
#else
  if (semctl(semid,IPC_RMID,0)) perror("semctl");
#endif
  //  if (index != 0) delete work;
}

void ThreadInfo::start() {
    ASSERT(index>=0);
#ifdef _THREAD_TRACE
    cout << "thread " << index << " start" << endl;
#endif
    signal();
}

void ThreadInfo::signal() {
    // unblock the thread so it can execute
#ifdef _WIN32
    if (!SetEvent(hEvent1)) {
        cerr << "error: SetEvent" << endl;
    }
#else
    struct sembuf sb = { 0, -1, IPC_NOWAIT };
    // decrement semaphore value:
    if (semop(semid,&sb,1)&& errno!=EAGAIN) 
       perror("semop: signal");
#endif
}

int ThreadInfo::wait() {
#ifdef _WIN32
  WaitForSingleObject(hEvent1,INFINITE);
#else
      struct sembuf sb = { 0, 0, 0 };
      // wait for semaphore value 0  
      if (semop(semid,&sb,1)) {
        if (errno == EINTR || errno == EIDRM) return 1;
        else {
          perror("semop: wait");
	}
      }
#endif
     return 0;
}

int ThreadInfo::wait(unsigned millisec) {
#ifdef _WIN32
  WaitForSingleObject(hEvent1,millisec);
#else
  struct timespec ts = { millisec/1000, (millisec%1000)*1000 };
      struct sembuf sb = { 0, 0, 0 };
      // wait for semaphore value 0  
      if (semtimedop(semid,&sb,1,&ts)) {
        if (errno == EINTR || errno == EIDRM) return 1;
        else if (errno != EAGAIN) {
          perror("semop: wait (timed)");
	}
      }
#endif
     return 0;
}

ThreadInfo::ThreadInfo(int i) 
 : state(Idle),
#ifdef _WIN32
   thread_id(NULL),
   work(NULL)
#else
   work(NULL),
   count(0)
#endif
   ,index(i)
{
#ifdef _THREAD_TRACE
  cout << "starting thread " << i << endl;
#endif
#ifdef _WIN32
      // initially the controller blocks the thread
      hEvent1 = CreateEvent(NULL,FALSE,FALSE,NULL);
      DWORD id;
      if (index == 0) {
	thread_id = GetCurrentThread();
      } else {
         thread_id = CreateThread(NULL,0,
            parkingLot,this,
            0,
            &id);
      }
#else
      semid = semget(IPC_PRIVATE,1,IPC_CREAT | 0777);
      if (semid == -1) {
	perror("semaphore creation failed");
        return;
      }
#ifdef _THREAD_TRACE
      cout << "semid = " << semid << endl;
#endif
      union semun val;
      // initialize semaphore to 1
      val.val = 1; 
      if (semctl(semid, 0, SETVAL, val)) {
        perror("semctl");
        return;
      }
      if (index == 0) {
	thread_id = pthread_self();
      }
      else {
        if (pthread_create(&thread_id, NULL, parkingLot, this)) {
  	   perror("thread creation failed");
	}
      }
#endif
}

ThreadPool::ThreadPool(SearchController *controller,int n) 
: nthreads(n) {
   data = new ThreadInfo *[n];
   LockInit(poolLock);
   for (int i = 0; i < n; i++) {
      data[i] = new ThreadInfo(i);
      data[i]->pool = this;
	  if (i==0) {
		 data[0]->work = new RootSearch(controller);
	     data[0]->work->ti = data[0];
	  }
	  else
		 data[i]->work = new Search(controller);
   }
}

ThreadPool::~ThreadPool() {
    shutDown();
}

void ThreadPool::shutDown() {
    // do not shut down thread 0 (main thread)
    for (int i = 1; i < nthreads; i++) {
       if (data[i]->state == ThreadInfo::Idle) {
          // set the thread to the terminating state that will force thread
	  // procedure exit
          data[i]->state = ThreadInfo::Terminating;
		  // unblock the thread
#ifdef _WIN32
          SetEvent(data[i]->hEvent1);
#else
       struct sembuf sb = { 0, -1, IPC_NOWAIT };
       // decrement semaphore value:
       if (semop(data[i]->semid,&sb,1)) perror("semop");
#endif
       }
       // wait for the thread to terminate
#ifdef _WIN32
        WaitForSingleObject(data[i]->thread_id,INFINITE);
#else
       void *value_ptr;
       pthread_join(data[i]->thread_id,&value_ptr);
#endif
       //delete data[i];
    }
    //delete data[0]; // main thread structure
    //delete [] data; 
}

ThreadInfo * ThreadPool::checkOut(Search *parent, NodeInfo *forNode,
  int ply, int depth) {
    Lock(poolLock);
    for (int i = 1; i < nthreads && i!=parent->ti->index; i++) {
#ifdef _WIN32
       if (data[i]->state == ThreadInfo::Idle) {
#else
       // verify the thread is available and is actually
       // at the synchronization point
       if (data[i]->state == ThreadInfo::Idle &&
           semctl(data[i]->semid,0,GETZCNT)>0) {
#endif
           // keep track of active threads for the calling node:
           Lock(parent->split_lock);
           if (!forNode->split) {
	          // this node didn't have a split point already, so make one,
              // if possible:
	          ASSERT(parent->split_stack_end>=parent->split_stack);
              if (parent->split_stack_end-parent->split_stack >= SPLIT_STACK_MAX_DEPTH)
		      {
		         Unlock(parent->split_lock);
                 Unlock(poolLock);
                 return NULL; // can't split this search instance any more
		      }
		      ASSERT(parent->split_stack);
	          forNode->split = parent->split_stack_end++;
              ASSERT(forNode->split->slaves.size()==0);
	       }
           ASSERT(forNode->split);
	       forNode->split->ply = ply;
	       forNode->split->depth = depth;
           forNode->split->index = forNode->num_try;
	       forNode->split->slaves.add(data[i]);
           Unlock(parent->split_lock);
 	       // obtained a thread, initialize its tracking object's state
           data[i]->state = ThreadInfo::Working;
           data[i]->parentNode = forNode;
           Unlock(poolLock);
#ifdef _THREAD_TRACE
           cout << "checkOut " << i << endl;
#endif
           return data[i]; 
       }
    }
    // no luck, no free threads
    Unlock(poolLock);
    return NULL;
}

void ThreadPool::checkIn(ThreadInfo *ti) {
#ifdef _THREAD_TRACE
    cout << "checkIn " << ti->index << endl;
#endif
    Lock(poolLock);
    Search *parent = ti->work->parent;
    Lock(parent->split_lock);
    Lock(ti->parentNode->nodeLock);
    // dissociate the thread from the parent
    ASSERT(ti->parentNode->split);
    ArasanSet<ThreadInfo *,Constants::MaxCPUs> &slaves = 
       ti->parentNode->split->slaves;
    ASSERT(slaves.exists(ti));
    // remove ti from the list of slave threads in the parent
    slaves.remove(ti);
    ASSERT(slaves.size()>=0 &&
           slaves.size()<MAX_CPUS);
    if (slaves.size() == 0) {
      // there are now no slave threads, so pop the split point stack
      ti->parentNode->split = NULL;
      parent->split_stack_end--;
      ASSERT(parent->split_stack_end>=parent->split_stack);
      // also signal the parent that it does not need to wait any more
      if (parent->ti->state == ThreadInfo::Waiting) {
        parent->ti->signal();
      }
    }
    Unlock(ti->parentNode->nodeLock);
    Unlock(parent->split_lock);
    Unlock(poolLock);
#ifdef _WIN32
    if (!ResetEvent(ti->hEvent1))
        cerr << "ResetEvent failed" << endl;
#else
    // increment semaphore value so we will wait
    // when entering loop again
    struct sembuf sb = { 0, 1, IPC_NOWAIT };
    if (semop(ti->semid,&sb,1)) perror("semop");
#endif
}

int ThreadPool::activeCount() {
    int count = 0;
    for (int i = 0; i < nthreads; i++) {
       if (data[i]->state == ThreadInfo::Working) ++count;
    }
    return count;
}

void ThreadPool::clearHashTables() {
  for (int i=1;i< nthreads;i++) {
    if (data[i]->work) data[i]->work->clearHashTables();
  }
}
