// Copyright 2005-2007 by Jon Dart. All Rights Reserved.

#ifndef _THREAD_POOL_H
#define _THREAD_POOL_H

#include "types.h"

class Search;
class SearchController;
struct NodeInfo;

class ThreadPool;

struct ThreadInfo {
   enum State { Idle, Working, Waiting, Terminating };
   ThreadInfo(int i);
   virtual ~ThreadInfo();
   State state;
   Search *work;
   NodeInfo *parentNode;
   ThreadPool *pool;
   THREAD thread_id;
#ifdef _WIN32
   HANDLE hEvent1;
#else
   int semid;  // semaphore id
   unsigned count;
#endif
   int index;
   void start();
   // wait for signal; return 1 if interrupted
   int wait();
   int wait(unsigned millisec);
   void signal();
};

class ThreadPool {
public:

   // create a thread pool with n threads
   ThreadPool(SearchController *,int n);

   virtual ~ThreadPool();

   ThreadInfo *mainThread() const {
     return data[0];
   }

   // obtain an idle thread if possible, returns 
   // non-null if successful
   ThreadInfo *checkOut(Search *,NodeInfo *,int ply,int depth);

   // return a thread to the pool
   void checkIn(ThreadInfo *);

   int activeCount();

   void clearHashTables();

private:
   void shutDown();

   LockDefine(poolLock);
   ThreadInfo **data;
   int nthreads;
};

#endif
